@@ -1,11 +1,8 @@
-/* -*- mode:C; indent-tabs-mode:t; tab-width:8; c-basic-offset:8; -*- */
-
-
 /* gxmessage - an xmessage clone using GTK2
  *
- * $Id: gxmessage.c,v 1.13 2004/04/15 11:25:46 trm Exp trm $
+ * $Id: gxmessage.c,v 1.18 2005/02/22 09:22:38 trm Exp trm $
  *
- * Copyright 2001 - 2004, Timothy Musson <trmusson@ihug.co.nz>
+ * Copyright Timothy Musson 2001 - 2005, <trmusson@ihug.co.nz>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -58,7 +55,7 @@ typedef struct _ButtonList ButtonList;
 
 
 
-struct GXMessage {
+struct {
 	gchar             *message_text;
 	gint              message_length;
 	ButtonList        *button_list;
@@ -69,13 +66,14 @@ struct GXMessage {
 	const gchar       *color_fg;
 	const gchar       *color_bg;
 	const gchar       *encoding;
+	const gchar       *entry_str;
 	gint              timeout;
 	guint             timeout_id;
 	gboolean          do_iconify;
 	gboolean          do_print;
-	gboolean          do_entry;
 	gboolean          do_buttons;
 	gboolean          do_borderless;
+	gboolean          do_focus;
 	GtkWrapMode       wrap_mode;
 	GtkWindowPosition window_position;
 	GtkWidget         *entry_widget;
@@ -86,7 +84,7 @@ struct GXMessage {
 struct _ButtonList {
 	gboolean          is_default;
 	gint              value;
-	gchar             *label_str;
+	const gchar       *label_str;
 	ButtonList        *prev;
 	ButtonList        *next;
 };
@@ -120,47 +118,55 @@ static struct Option option[] = { /* min_len, requires_arg, opt_str */
 	{2, FALSE, "rv"},
 	{2, FALSE, "reverse"},
 	{2, TRUE,  "selectionTimeout"},
+	{2, FALSE, "synchronous"},
 	{2, TRUE,  "xnllanguage"},
 	{2, FALSE, "borderless"},
 	{1, FALSE, "wrap"},
 	{3, TRUE,  "encoding"},
+	{3, FALSE, "nofocus"},
+	{6, TRUE,  "entrytext"},
+	{3, FALSE, "entry"},
 	{1, FALSE, "?"},
 	{1, FALSE, "help"},
-	{1, FALSE, "version"},
-	{1, FALSE, "entry"}
+	{1, FALSE, "version"}
 };
-#define OPT_BUTTONS           0
-#define OPT_CENTER            1
-#define OPT_DEFAULT           2
-#define OPT_FILE              3
-#define OPT_NEARMOUSE         4
-#define OPT_PRINT             5
-#define OPT_TIMEOUT           6
-#define OPT_FN                7
-#define OPT_FONT              8
-#define OPT_GEOMETRY          9
-#define OPT_TITLE             10
-#define OPT_BG                11
-#define OPT_FG                12
-#define OPT_BD                13
-#define OPT_BW                14
-#define OPT_ICONIC            15
-#define OPT_XRM               16
-#define OPT_RV                17
-#define OPT_REVERSE           18
-#define OPT_SELECTIONTIMEOUT  19
-#define OPT_XNLLANGUAGE       20
-#define OPT_BORDERLESS        21
-#define OPT_WRAP              22
-#define OPT_ENCODING          23
-#define OPT_HELP_Q            24
-#define OPT_HELP              25
-#define OPT_VERSION           26
-#define OPT_ENTRY             27
-#define N_OPTS                28
 
-#define OPT_IS_MISSING_ARG    -1
-#define OPT_IS_UNKNOWN        -2
+enum {
+	OPT_IS_UNKNOWN = -2,
+	OPT_IS_MISSING_ARG,
+	OPT_BUTTONS,
+	OPT_CENTER,
+	OPT_DEFAULT,
+	OPT_FILE,
+	OPT_NEARMOUSE,
+	OPT_PRINT,
+	OPT_TIMEOUT,
+	OPT_FN,
+	OPT_FONT,
+	OPT_GEOMETRY,
+	OPT_TITLE,
+	OPT_BG,
+	OPT_FG,
+	OPT_BD,
+	OPT_BW,
+	OPT_ICONIC,
+	OPT_XRM,
+	OPT_RV,
+	OPT_REVERSE,
+	OPT_SELECTIONTIMEOUT,
+	OPT_SYNCHRONOUS,
+	OPT_XNLLANGUAGE,
+	OPT_BORDERLESS,
+	OPT_WRAP,
+	OPT_ENCODING,
+	OPT_FOCUS,
+	OPT_ENTRYTEXT,
+	OPT_ENTRY,
+	OPT_HELP_Q,
+	OPT_HELP,
+	OPT_VERSION,
+	N_OPTS
+};
 
 
 
@@ -171,20 +177,11 @@ void prog_cleanup (void);
 ButtonList*
 button_first (ButtonList *button)
 {
-	if (button == NULL) return NULL;
-
-	while (button->prev != NULL) button = button->prev;
-	return button;
-}
-
-
-
-ButtonList*
-button_new (void)
-{
-	ButtonList *button = g_new0 (ButtonList, 1);
-
-	if (button == NULL) prog_cleanup ();
+	if (button != NULL) {
+		while (button->prev != NULL) {
+			button = button->prev;
+		}
+	}
 	return button;
 }
 
@@ -193,14 +190,13 @@ button_new (void)
 void
 button_free (ButtonList *button)
 {
-	ButtonList *button_next;
+	ButtonList *next;
 
 	button = button_first (button);
 	while (button != NULL) {
-		button_next = button->next;
-		g_free (button->label_str);
+		next = button->next;
 		g_free (button);
-		button = button_next;
+		button = next;
 	}
 }
 
@@ -219,29 +215,29 @@ button_append (ButtonList *button, ButtonList *new_button)
 
 
 ButtonList*
-button_new_from_str (const gchar *str, gint default_value, gint len)
+button_new_from_str (gchar *str, gint default_value, gint len)
 {
-	/*
-	 * Given a string containing "label" or "label:value",
-	 * fill in a new button structure and return it.
+	/* Given a string containing "label" or "label:value", fill in a
+	 * new button structure and return it.
 	 */
-	ButtonList *button = button_new ();
+
+	ButtonList *button = g_new0 (ButtonList, 1);
 	gboolean value_from_str = FALSE;
 	gint i = len;
 
 
 	/* parse "label:value" format in str */
+	str[len] = '\0';
 	while (i != 0 && !value_from_str) value_from_str = str[--i] == ':';
 	if (value_from_str) {
 		/* get label and value from str */
-		button->label_str = g_strndup (str, i);
-		if (button->label_str == NULL) prog_cleanup ();
-		button->value = atoi (&str[i+1]);
+		str[i] = '\0';
+		button->label_str = str;
+		button->value = atoi (&str[i + 1]);
 	}
 	else {
 		/* get label from str, value from default_value */
-		button->label_str = g_strndup (str, len);
-		if (button->label_str == NULL) prog_cleanup ();
+		button->label_str = str;
 		button->value = default_value;
 	}
 
@@ -256,14 +252,13 @@ button_new_from_str (const gchar *str, gint default_value, gint len)
 
 
 ButtonList*
-button_list_from_str (const gchar *str)
+button_list_from_str (gchar *str)
 {
-	/*
-	 * Given a string containing a comma-separated list of
-	 * "label" and/or "label:value", build and return a list
-	 * of buttons. The return points to the last button in
-	 * the list.
+	/* Given a string containing a comma-separated list of "label" and/or
+	 * "label:value", build and return a list of buttons. The return points
+	 * to the last button in the list.
 	 */
+
 	ButtonList *button = NULL;
 	ButtonList *b;
 	gint len;
@@ -296,12 +291,8 @@ button_list_from_str (const gchar *str)
 void
 button_set_default (ButtonList *button, const gchar *str)
 {
-	/*
-	 * Make button->is_default TRUE for any button whose label
-	 * matches str. (If there's more than one, only the last
-	 * one found will actually be used as the default - that
-	 * happens automatically in window_create.)
-	 */
+	/* Make button->is_default TRUE for each button whose label matches str */
+
 	if (str == NULL) return;
 
 	button = button_first (button);
@@ -314,15 +305,14 @@ button_set_default (ButtonList *button, const gchar *str)
 
 
 gint
-my_get_opt (const gchar *str, gboolean is_last)
+my_get_opt (const gchar *str, gboolean not_last)
 {
-	/*
-	 * Try to identify the command line option in str,
-	 * returning a unique option/error code. The
-	 * is_last variable specifies whether or not the
-	 * current option was followed by something else
-	 * (a value or another option) on the command line.
+	/* Try to identify the command line option in str, returning a unique
+	 * option/error code. The not_last variable specifies whether current
+	 * option was followed by something else (a value or another option)
+	 * on the command line.
 	 */
+
 	gint opt;
 	gint len;
 	gint offset = 1;
@@ -339,8 +329,9 @@ my_get_opt (const gchar *str, gboolean is_last)
 		}
 		if (len > 0) {
 			for (opt = 0; opt < N_OPTS; opt++) {
-				if (len >= option[opt].min_len && strncmp (&str[offset], option[opt].opt_str, len) == 0) {
-					if (!option[opt].requires_arg || is_last) {
+				if (len >= option[opt].min_len &&
+				    strncmp (&str[offset], option[opt].opt_str, len) == 0) {
+					if (!option[opt].requires_arg || not_last) {
 						return opt;
 					}
 					else {
@@ -371,7 +362,7 @@ cb_button_clicked (GtkWidget *widget, gpointer data)
 	if (gx.do_print) {
 		g_print ("%s\n", ((ButtonList*)data)->label_str);
 	}
-	else if (gx.do_entry) {
+	else if (gx.entry_str != NULL) {
 		g_print ("%s\n", gtk_entry_get_text (GTK_ENTRY(gx.entry_widget)));
 	}
 	gtk_main_quit ();
@@ -427,10 +418,9 @@ gboolean
 geometry_parse_string (const gchar *geometry,
                        gint *x, gint *y, gint *w, gint *h)
 {
-	/*
-	 * Parse geometry string into x, y, w and h. Return TRUE
-	 * if successful, else FALSE. (This code is taken straight
-	 * from the GNOME 1.x gnome-geometry.c)
+	/* Parse geometry string into x, y, w and h. Return TRUE if successful,
+	 * else FALSE. (This function and geometry_get_number are taken from
+	 * the GNOME 1.x gnome-geometry.c)
 	 */
 
 	gint subtract;
@@ -514,7 +504,6 @@ window_create (void)
 	GtkWidget  *window;
 	GtkWidget  *vbox, *vbox2;
 	GtkWidget  *scroller;
-	GtkWidget  *sep = NULL;
 	GtkWidget  *button_box = NULL;
 	GtkWidget  *widget = NULL;
 	GtkWidget  *message_widget;
@@ -523,16 +512,15 @@ window_create (void)
 	PangoFontDescription *font_desc;
 	GtkRequisition size_req;
 	GtkTextBuffer *buf;
-	gint win_w, win_h, scr_w, scr_h, max_default_w, max_default_h;
-	gint tmp_w, tmp_h, tmp_x, tmp_y;
+	gint win_w, win_h;
 
 
 	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 
 
-	g_signal_connect (GTK_OBJECT(window),
+	g_signal_connect (G_OBJECT(window),
 	                  "destroy",
-	                  GTK_SIGNAL_FUNC(cb_window_destroy),
+	                  G_CALLBACK(cb_window_destroy),
 	                  NULL);
 
 
@@ -547,14 +535,19 @@ window_create (void)
 	gtk_window_set_icon_from_file (GTK_WINDOW(window), FNAME_ICON, NULL);
 
 	if (gx.do_borderless) {
-		gtk_window_set_decorated (GTK_WINDOW(window), 0);
+		gtk_window_set_decorated (GTK_WINDOW(window), FALSE);
 	}
+#if _GTK_2_0
+#else
+	/* requires GTK >= 2.4.0 */
+	gtk_window_set_accept_focus (GTK_WINDOW(window), gx.do_focus);
+#endif
 
 
 	/* window contents */
-	gtk_container_set_border_width (GTK_CONTAINER(window), 6);
+	gtk_container_set_border_width (GTK_CONTAINER(window), 12);
 
-	vbox = gtk_vbox_new (FALSE, 0);
+	vbox = gtk_vbox_new (FALSE, 12);
 	gtk_container_add (GTK_CONTAINER(window), vbox);
 
 	vbox2 = gtk_vbox_new (FALSE, 0);
@@ -572,15 +565,8 @@ window_create (void)
 	/* the message */
 	message_widget = gtk_text_view_new ();
 	gtk_text_view_set_editable (GTK_TEXT_VIEW(message_widget), FALSE);
-	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW(message_widget), FALSE);
+	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW(message_widget), TRUE);
 	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(message_widget), gx.wrap_mode);
-	gtk_text_view_set_justification (GTK_TEXT_VIEW(message_widget), GTK_JUSTIFY_LEFT);
-	gtk_text_view_set_left_margin (GTK_TEXT_VIEW(message_widget), 2);
-	gtk_text_view_set_right_margin (GTK_TEXT_VIEW(message_widget), 2);
-	gtk_text_view_set_pixels_above_lines (GTK_TEXT_VIEW(message_widget), 2);
-	gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW(message_widget), 2);
-	gtk_text_view_set_pixels_inside_wrap (GTK_TEXT_VIEW(message_widget), 0);
-	gtk_text_view_set_indent (GTK_TEXT_VIEW(message_widget), 0);
 
 	if (gx.font_str != NULL) {
 		font_desc = pango_font_description_from_string (gx.font_str);
@@ -598,26 +584,29 @@ window_create (void)
 	gtk_text_buffer_set_text (buf, gx.message_text, gx.message_length);
 
 	if (gx.color_fg != NULL) {
-		gdk_color_parse (gx.color_fg, &color);
-		gtk_widget_modify_text (message_widget, GTK_STATE_NORMAL, &color);
+		if (gdk_color_parse (gx.color_fg, &color)) {
+			gtk_widget_modify_text (message_widget, GTK_STATE_NORMAL, &color);
+		}
 	}
 	if (gx.color_bg != NULL) {
-		gdk_color_parse (gx.color_bg, &color);
-		gtk_widget_modify_base (message_widget, GTK_STATE_NORMAL, &color);
+		if (gdk_color_parse (gx.color_bg, &color)) {
+			gtk_widget_modify_base (message_widget, GTK_STATE_NORMAL, &color);
+		}
 	}
 
 
 	/* text entry */
-	if (gx.do_entry) {
+	if (gx.entry_str != NULL) {
 		gx.entry_widget = gtk_entry_new ();
 		gtk_editable_set_editable (GTK_EDITABLE(gx.entry_widget), TRUE);
+		gtk_entry_set_text (GTK_ENTRY(gx.entry_widget), gx.entry_str);
 		gtk_box_pack_start (GTK_BOX(vbox), gx.entry_widget, FALSE, FALSE, 5);
 		gtk_widget_grab_focus (gx.entry_widget);
 		if (!gx.do_buttons) {
 			/* allow hitting <RETURN> to close the window */
-			g_signal_connect (GTK_OBJECT(gx.entry_widget),
+			g_signal_connect (G_OBJECT(gx.entry_widget),
 			                  "activate",
-			                  GTK_SIGNAL_FUNC(cb_entry_activated),
+			                  G_CALLBACK(cb_entry_activated),
 			                  (gpointer)0);
 		}
 	}
@@ -626,20 +615,40 @@ window_create (void)
 	/* add buttons */
 	if (gx.do_buttons) {
 
-		sep = gtk_hseparator_new ();
-		gtk_box_pack_start (GTK_BOX(vbox), sep, FALSE, FALSE, 5);
-
 		button_list = button_first (gx.button_list);
 
 		button_box = gtk_hbutton_box_new ();
+		gtk_button_box_set_layout (GTK_BUTTON_BOX(button_box),
+		                           GTK_BUTTONBOX_END);
+		gtk_box_set_spacing (GTK_BOX(button_box), 6);
 		gtk_box_pack_end (GTK_BOX(vbox), button_box, FALSE, FALSE, 0);
 
 		while (button_list != NULL) {
 
-			widget = gtk_button_new_with_label (button_list->label_str);
-			g_signal_connect (GTK_OBJECT(widget),
+			if (g_str_has_prefix (button_list->label_str, "GTK_STOCK_")) {
+				gchar *s;
+				gchar *p;
+				p = g_ascii_strdown (button_list->label_str + 10, -1);
+				s = p;
+				while (*s != '\0') {
+					if (*s == '_') *s = '-';
+					s++;
+				}
+				s = g_strconcat ("gtk-", p, NULL);
+				widget = gtk_button_new_from_stock (s);
+				g_free (s);
+				g_free (p);
+				if (widget == NULL) {
+					widget = gtk_button_new_with_label (button_list->label_str);
+				}
+			}
+			else {
+				widget = gtk_button_new_with_mnemonic (button_list->label_str);
+			}
+
+			g_signal_connect (G_OBJECT(widget),
 			                  "clicked",
-			                  GTK_SIGNAL_FUNC(cb_button_clicked),
+			                  G_CALLBACK(cb_button_clicked),
 			                  (gpointer)button_list);
 			gtk_box_pack_start (GTK_BOX(button_box), widget, FALSE, FALSE, 0);
 
@@ -648,65 +657,28 @@ window_create (void)
 			}
 
 			button_list = button_list->next;
-
 		}
 
 	}
 
 
-	/*
-	 * figure out window geometry
-	 */
+	/* window geometry */
 
-	/* start by getting the width and height of the GtkTextView's contents */
+	win_w = 330;  /* XXX */
+	win_h = 90;   /* XXX */
 
-	gtk_widget_realize (message_widget);
-#if 0
-	/* FIXME: this only gets the _displayed_ geometry, but I need the full
-	 * size of the GtkTextView widget's "contents" */
-	/* gdk_window_get_geometry (GDK_WINDOW(message_widget->window), NULL, NULL, &win_w, &win_h, NULL); */
-	{
-		GtkRequisition req;
-		gtk_widget_size_request (message_widget, &req);
-		g_print ("%d %d\n", req.width, req.height);
-		win_w = req.width;
-		win_h = req.height;
-	}
-#else
-	/* FIXME: these are temporary defaults */
-	win_w = 300;
-	win_h = 80;
-#endif
-
-	if (gx.do_entry) {
-		/* get entry height + padding */
+	if (gx.entry_str != NULL) {
 		gtk_widget_realize (gx.entry_widget);
 		gtk_widget_size_request (gx.entry_widget, &size_req);
 		win_h = win_h + size_req.height + 10;
 	}
 	if (gx.do_buttons) {
-		/* get button height + separator height + padding */
 		gtk_widget_size_request (button_box, &size_req);
 		win_h = win_h + size_req.height;
-		gtk_widget_size_request (sep, &size_req);
-		win_h = win_h + size_req.height + 10;
 	}
 
-	/* a bit of extra padding */
-	win_w = win_w + 30;
-	win_h = win_h + 10;
-
-	/* no bigger than 75% screen width x 75% screen height by default */
-	scr_w = gdk_screen_width ();
-	scr_h = gdk_screen_height ();
-	max_default_w = scr_w * 0.75;
-	max_default_h = scr_h * 0.75;
-	if (win_w > max_default_w) win_w = max_default_w;
-	if (win_h > max_default_h) win_h = max_default_h;
-
-
-	/* if a geometry string was given, try to use it for width and height */
 	if (gx.geom_str != NULL) {
+		gint tmp_x, tmp_y, tmp_w, tmp_h;
 		if (geometry_parse_string (gx.geom_str, &tmp_x, &tmp_y, &tmp_w, &tmp_h)) {
 			if (win_w != -1) {
 				win_w = tmp_w;
@@ -715,15 +687,12 @@ window_create (void)
 		}
 	}
 
-
-	/* set height and position */
 	gtk_window_set_default_size (GTK_WINDOW(window), win_w, win_h);
 	gtk_window_set_position (GTK_WINDOW(window), gx.window_position);
 
 	/* open the window */
 	gtk_widget_show_all (window);
 
-
 	/* begin timeout */
 	if (gx.timeout != 0) {
 		gx.timeout_id = gtk_timeout_add (1000, (GtkFunction)cb_timeout,
@@ -741,10 +710,9 @@ read_stdin (void)
 	gint ch;
 
 	text = g_string_new ("");
-	if (text == NULL) prog_cleanup ();
 
 	while ( (ch = getc (stdin)) != EOF ) {
-		if (g_string_append_c (text, ch) == NULL) prog_cleanup ();
+		g_string_append_c (text, ch);
 	}
 	str = text->str;
 	g_string_free (text, FALSE);
@@ -788,56 +756,34 @@ message_to_utf8 (const gchar *str)
 void
 option_kludge (gint *argc, gchar ***argv)
 {
-	/*
-	 * Change -display, -name and -synchronous to
-	 * --display, --name and --sync, ready for gtk_init.
-	 * (This function is based on code in GTK 1.2)
-	 */
+	/* Vet argv before gtk_init sees it */
 
-	gint i, j, k, len;
-
-
-	for (i = 1; i < *argc;) {
+	gint i, len;
 
+	for (i = 1; i < *argc; i++) {
 		len = strlen ((*argv)[i]);
-
 		if (len > 2) {
-
 			if (strncmp ("-display", (*argv)[i], len) == 0) {
 				(*argv)[i] = "--display";
-				if ((i + 1) < *argc && (*argv)[i + 1]) {
-					i += 1;
-				}
+				if (i + 1 < *argc && (*argv)[i + 1] != NULL) i++;
 			}
 			else if (strncmp ("-name", (*argv)[i], len) == 0) {
 				(*argv)[i] = "--name";
-				if ((i + 1) < *argc && (*argv)[i + 1]) {
-					i += 1;
-				}
-			}
-			else if (strncmp ("-synchronous", (*argv)[i], len) == 0) {
-				(*argv)[i] = "--sync";
+				if (i + 1 < *argc && (*argv)[i + 1] != NULL) i++;
 			}
 		}
-
-		i += 1;
-
 	}
 
-	for (i = 1; i < *argc; i++) {
-		for (k = i; k < *argc; k++) {
-			if ((*argv)[k] != NULL) {
-				break;
-			}
-		}
-		if (k > i) {
-			k -= i;
-			for (j = i + k; j < *argc; j++) {
-				(*argv)[j-k] = (*argv)[j];
-			}
-			*argc -= k;
+#if _GTK_2_6
+	/* work around the removal of '-' by gtk_init >= 2.6.0 */
+	for (i = 2; i < *argc; i++) {
+		if (strcmp ((*argv)[i], "-") == 0 &&
+		    (strcmp ((*argv)[i - 1], "-file") == 0 ||
+		     strcmp ((*argv)[i - 1], "--file") == 0)) {
+			(*argv)[i] = "gxmsg:stdin";
 		}
 	}
+#endif
 }
 
 
@@ -867,8 +813,10 @@ usage (void)
 	        "  -title TITLE           Set window title to TITLE\n\n"));
 	g_print(_("%s options:\n" \
 	        "  -borderless            Try to open window without border decoration\n" \
+	        "  -nofocus               Don't focus the window when it opens\n" \
 	        "  -encoding CHARSET      Expect CHARSET as the message encoding\n" \
 	        "  -entry                 Prompt for text to be sent to stdout\n" \
+	        "  -entrytext TEXT        Same as -entry, but with TEXT as default text\n" \
 	        "  -wrap                  Wrap lines of text to fit window width\n" \
 	        "  -help | -?             Show this usage information\n" \
 	        "  -version               Show gxmessage version information\n\n"), PACKAGE);
@@ -878,32 +826,6 @@ usage (void)
 
 
 
-void
-gxmessage_init (void)
-{
-	gx.message_text    = NULL;
-	gx.message_length  = 0;
-	gx.button_list     = NULL;
-	gx.default_str     = NULL;
-	gx.title_str       = NULL;
-	gx.geom_str        = NULL;
-	gx.font_str        = NULL;
-	gx.color_fg        = NULL;
-	gx.color_bg        = NULL;
-	gx.encoding        = NULL;
-	gx.timeout         = 0;
-	gx.timeout_id      = 0;
-	gx.do_print        = FALSE;
-	gx.do_iconify      = FALSE;
-	gx.do_buttons      = FALSE;
-	gx.do_borderless   = FALSE;
-	gx.wrap_mode       = GTK_WRAP_NONE;
-	gx.window_position = GTK_WIN_POS_NONE;
-	gx.exit_code       = 1;
-}
-
-
-
 void
 prog_cleanup (void)
 {
@@ -924,9 +846,14 @@ int
 main (gint argc, gchar *argv[])
 {
 	GString *gstr = NULL;
-	gboolean input_is_file = FALSE;
 	gchar *ch = NULL, *tmpstr;
+	const gchar *fname = NULL;
 	gint opt, arg = 1;
+	gchar bu_default[] = "okay:0";
+
+	/* The default "okay:0" string is intentionally hard-wired, to avoid
+	 * breaking scripts that make use of xmessage's -print option.
+	 * It must not be changed or gettextize'd. */
 
 
 #ifdef ENABLE_NLS
@@ -937,11 +864,15 @@ main (gint argc, gchar *argv[])
 
 	option_kludge (&argc, &argv);
 	gtk_init (&argc, &argv);
-	gxmessage_init ();
+
+	gx.exit_code       = 1;
+	gx.do_focus        = TRUE;
+	gx.wrap_mode       = GTK_WRAP_NONE;
+	gx.window_position = GTK_WIN_POS_NONE;
 
 
 	while (arg < argc) {
-		opt = my_get_opt (argv[arg], arg+1 < argc);
+		opt = my_get_opt (argv[arg], arg + 1 < argc);
 		switch (opt) {
 		case OPT_HELP :
 		case OPT_HELP_Q :
@@ -953,6 +884,7 @@ main (gint argc, gchar *argv[])
 			exit (0);
 			break;
 		case OPT_ENTRY :
+		case OPT_ENTRYTEXT :
 			if (gx.do_print) {
 				g_printerr (_("%s: can't have both -entry and -print\n"), PACKAGE);
 				prog_cleanup ();
@@ -961,7 +893,12 @@ main (gint argc, gchar *argv[])
 				g_printerr (_("%s: can't have both -entry and -timeout\n"), PACKAGE);
 				gx.timeout = 0;
 			}
-			gx.do_entry = TRUE;
+			if (opt == OPT_ENTRY) {
+				gx.entry_str = "";
+			}
+			else {
+				gx.entry_str = argv[++arg];
+			}
 			break;
 		case OPT_BUTTONS :
 			button_free (gx.button_list);
@@ -974,26 +911,12 @@ main (gint argc, gchar *argv[])
 			gx.default_str = argv[++arg];
 			break;
 		case OPT_FILE :
-			if (input_is_file) {
-				g_printerr (_("%s: only one -file allowed\n"), PACKAGE);
-				prog_cleanup ();
-			}
-			arg++;
-			input_is_file = TRUE;
 			if (gstr != NULL) {
 				g_printerr (_("%s: can't get message from both -file and command line\n"), PACKAGE);
 				prog_cleanup ();
 			}
-			if (strcmp ("-", argv[arg]) == 0) {
-				tmpstr = read_stdin ();
-			}
-			else if (!g_file_get_contents (argv[arg], &tmpstr, NULL, NULL)) {
-				g_printerr (_("%s: unable to read file\n"), PACKAGE);
-				prog_cleanup ();
-			}
-			gx.message_text = message_to_utf8 (tmpstr);
-			gx.message_length = strlen (gx.message_text);
-			g_free (tmpstr);
+			arg++;
+			fname = argv[arg];
 			break;
 		case OPT_NEARMOUSE :
 			/* -center takes priority over -nearmouse */
@@ -1002,7 +925,7 @@ main (gint argc, gchar *argv[])
 			}
 			break;
 		case OPT_PRINT :
-			if (gx.do_entry) {
+			if (gx.entry_str != NULL) {
 				g_printerr (_("%s: can't have both -entry and -print\n"), PACKAGE);
 				prog_cleanup ();
 			}
@@ -1017,7 +940,7 @@ main (gint argc, gchar *argv[])
 				/* xmessage doesn't complain here, just disables timeout */
 				gx.timeout = 0;
 			}
-			if (gx.do_entry) {
+			if (gx.entry_str != NULL) {
 				g_printerr (_("%s: can't have both -entry and -timeout\n"), PACKAGE);
 				gx.timeout = 0;
 			}
@@ -1034,6 +957,7 @@ main (gint argc, gchar *argv[])
 			break;
 		case OPT_RV :
 		case OPT_REVERSE :
+		case OPT_SYNCHRONOUS :
 			/* not implemented - ignore */
 			break;
 		case OPT_BG :
@@ -1062,44 +986,59 @@ main (gint argc, gchar *argv[])
 		case OPT_ENCODING :
 			gx.encoding = argv[++arg];
 			break;
+		case OPT_FOCUS:
+			gx.do_focus = FALSE;
+			break;
 		case OPT_IS_MISSING_ARG :
 			/* in this case, xmessage treats the "option" as normal text */
 		case OPT_IS_UNKNOWN :
 		default:
-			if (input_is_file) {
+			if (fname != NULL) {
 				g_printerr (_("%s: can't get message from both -file and command line\n"), PACKAGE);
 				prog_cleanup ();
 			}
 			if (gstr == NULL) {
 				gstr = g_string_new ("");
-				if (gstr == NULL) prog_cleanup ();
 			}
 			else {
 				gstr = g_string_append_c (gstr, ' ');
-				if (gstr == NULL) prog_cleanup ();
 			}
 			gstr = g_string_append (gstr, argv[arg]);
-			if (gstr == NULL) prog_cleanup ();
 			break;
 		}
 		arg++;
 	}
 
-
-	if (gstr != NULL) {
+	if (fname != NULL) {
+#if _GTK_2_6
+		if (strcmp ("gxmsg:stdin", fname) == 0) {
+#else
+		if (strcmp ("-", fname) == 0) {
+#endif
+			tmpstr = read_stdin ();
+		}
+		else if (!g_file_get_contents (fname, &tmpstr, NULL, NULL)) {
+			g_printerr (_("%s: unable to read file\n"), PACKAGE);
+			prog_cleanup ();
+		}
+		gx.message_text = message_to_utf8 (tmpstr);
+		gx.message_length = strlen (gx.message_text);
+		g_free (tmpstr);
+	}
+	else if (gstr != NULL) {
 		gx.message_text = message_to_utf8 (gstr->str);
 		gx.message_length = strlen (gx.message_text);
 		g_string_free (gstr, TRUE);
 	}
-
-	if (gx.message_text == NULL) {
+	else {
 		g_printerr (_("%s: message text is required\n"), PACKAGE);
 		g_printerr (_("Try `%s --help' for more information\n"), PACKAGE);
 		prog_cleanup ();
 	}
 
 	if (gx.button_list == NULL) {
-		gx.button_list = button_append (NULL, button_new_from_str ("okay:0", 0, 6));
+		gx.button_list = button_append (NULL,
+		    button_new_from_str (bu_default, 0, strlen (bu_default)));
 	}
 
 
@@ -1113,3 +1052,4 @@ main (gint argc, gchar *argv[])
 	prog_cleanup ();
 	return 0;
 }
+
